"""
This module is an example of a barebones QWidget plugin for napari

It implements the Widget specification.
see: https://napari.org/plugins/guides.html?#widgets

Replace code below according to your needs.
"""
from enum import Enum
from functools import partial

import dask.array as da
import numpy as np
from magicgui import magic_factory
from napari.layers import Labels
from napari.utils import progress
from qtpy.QtWidgets import (
    QComboBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)
from skimage.filters import (
    threshold_isodata,
    threshold_li,
    threshold_otsu,
    threshold_triangle,
    threshold_yen,
)
from skimage.measure import label


class Threshold(Enum):
    isodata = partial(threshold_isodata)
    li = partial(threshold_li)
    otsu = partial(threshold_otsu)
    triangle = partial(threshold_triangle)
    yen = partial(threshold_yen)


class SegmentationDiffHighlight(QWidget):
    def __init__(self, napari_viewer):
        super().__init__()
        self.viewer = napari_viewer
        self.setLayout(QVBoxLayout())
        self.highlight_btn = QPushButton("Highlight Differences")
        self.highlight_btn.clicked.connect(self._compute_differences)
        self.layer_combos = []
        self.gt_layer_combo = self.add_labels_combo_box("Ground Truth Layer")
        self.seg_layer_combo = self.add_labels_combo_box("Segmentation Layer")
        self.viewer.layers.events.inserted.connect(self._reset_layer_options)
        self.viewer.layers.events.removed.connect(self._reset_layer_options)
        self.layout().addWidget(self.highlight_btn)
        self.layout().addStretch()

    def add_labels_combo_box(self, label_text):
        combo_row = QWidget()
        combo_row.setLayout(QHBoxLayout())
        combo_row.layout().setContentsMargins(0, 0, 0, 0)
        new_combo_label = QLabel(label_text)
        combo_row.layout().addWidget(new_combo_label)
        new_layer_combo = QComboBox(self)
        new_layer_combo.addItems(
            [
                layer.name
                for layer in self.viewer.layers
                if isinstance(layer, Labels)
            ]
        )
        combo_row.layout().addWidget(new_layer_combo)
        self.layer_combos.append(new_layer_combo)
        self.layout().addWidget(combo_row)
        return new_layer_combo

    def _compute_differences(self):
        gt_layer = self.viewer.layers[self.gt_layer_combo.currentText()]
        seg_layer = self.viewer.layers[self.seg_layer_combo.currentText()]
        truth_foreground = da.where(gt_layer.data != 0, 1, 0)
        seg_foreground = da.where(seg_layer.data != 0, 1, 0)
        for i in range(len(truth_foreground)):
            if np.count_nonzero(truth_foreground[i]) == 0:
                seg_foreground[i] = da.zeros(truth_foreground[i].shape)
        diff = da.where(truth_foreground != seg_foreground, 1, 0)
        gt_layer.visible = False
        seg_layer.visible = False
        self.viewer.add_labels(diff, name="seg_gt_diff", color={1: "#fca503"})

    def _reset_layer_options(self, event):
        for combo in self.layer_combos:
            combo.clear()
            combo.addItems(
                [
                    layer.name
                    for layer in self.viewer.layers
                    if isinstance(layer, Labels)
                ]
            )


@magic_factory
def segment_by_threshold(
    img_layer: "napari.layers.Image", threshold: Threshold
) -> "napari.types.LayerDataTuple":
    with progress(total=0):
        # need to use threshold.value to get the function from the enum member
        threshold_val = threshold.value(img_layer.data.compute())
        binarised_im = img_layer.data > threshold_val
        seg_labels = da.from_array(label(binarised_im))

    seg_layer = (seg_labels, {"name": f"{img_layer.name}_seg"}, "labels")

    return seg_layer



import dask.array as da
import pytest
from napari.layers import Image, Labels
from qtpy.QtCore import Qt

from tracking_challenge_demo import (
    SegmentationDiffHighlight,
    Threshold,
    segment_by_threshold,
)


@pytest.fixture
def im_layer():
    return Image(da.random.random((5, 100, 100)), name="im")


@pytest.fixture
def labels_layer():
    return Labels(da.random.randint(0, 255, (5, 100, 100)), name="lab")


def test_segment_widg_returns_layer(im_layer, labels_layer):
    widg = segment_by_threshold()

    retval = widg(im_layer, Threshold.triangle)
    assert isinstance(retval[0], da.Array)
    assert retval[1]["name"] == "im_seg"
    assert retval[2] == "labels"


def test_highlight_widg_populates_layers(
    make_napari_viewer, labels_layer, im_layer
):
    viewer = make_napari_viewer()
    widg = SegmentationDiffHighlight(viewer)

    assert widg.gt_layer_combo.count() == 0
    assert widg.seg_layer_combo.count() == 0

    viewer.add_layer(labels_layer)
    viewer.add_layer(im_layer)
    assert widg.gt_layer_combo.count() == 1
    assert widg.gt_layer_combo.currentText() == "lab"
    assert widg.seg_layer_combo.count() == 1
    assert widg.seg_layer_combo.currentText() == "lab"


def test_highlight_widg_computes_difference(
    make_napari_viewer, labels_layer, qtbot
):
    viewer = make_napari_viewer()
    widg = SegmentationDiffHighlight(viewer)
    viewer.add_layer(labels_layer)
    assert len(viewer.layers) == 1

    qtbot.mouseClick(widg.highlight_btn, Qt.MouseButton.LeftButton)

    assert len(viewer.layers) == 2



"""
This module is an example of a barebones numpy reader plugin for napari.

It implements the Reader specification, but your plugin may choose to
implement multiple readers or even other plugin contributions. see:
https://napari.org/plugins/guides.html?#readers
"""
import glob
import os
import re
import warnings
from pathlib import Path

import dask
import dask.array as da
import tifffile
from napari.utils import progress

SEQ_REGEX = r"(.*)/([0-9]{2,})$"
GT_REGEX = r"(.*)/([0-9]{2,})(_(?:GT|AUTO))/SEG$"

SEQ_TIF_REGEX = rf'{SEQ_REGEX[:-1]}/t([0-9]{{3}}){"."}tif$'
GT_TIF_REGEX = rf'{GT_REGEX[:-1]}/(?:man_)?seg([0-9]{{3}}){"."}tif$'


def napari_get_reader(path):
    path = os.path.abspath(path)
    if not os.path.isdir(path):
        return None
    is_gt = re.match(GT_REGEX, path)
    if not is_gt:
        return None
    all_tifs = glob.glob(path + "/*.tif")
    if not all_tifs:
        return None
    is_gt_tifs = all([re.match(GT_TIF_REGEX, pth) for pth in all_tifs])
    if not is_gt_tifs:
        return None

    return reader_function


def reader_function(path):
    path = os.path.normpath(path)
    gt_match = re.match(GT_REGEX, path)

    parent_dir_pth = Path(path).parent.parent.absolute()
    seq_number = gt_match.groups()[1]
    sister_sequence_pth = os.path.join(parent_dir_pth, seq_number)

    n_frames = None
    if not os.path.exists(sister_sequence_pth):
        warnings.warn(
            f"Can't find image for ground truth at {path}. Reading without knowing number of frames..."
        )
    else:
        latest_tif_pth = sorted(glob.glob(sister_sequence_pth + "/*.tif"))[-1]
        n_frames = (
            int(re.match(SEQ_TIF_REGEX, latest_tif_pth).groups()[-1]) + 1
        )

    all_tifs = sorted(
        pth
        for pth in glob.glob(path + "/*.tif")
        if re.match(GT_TIF_REGEX, pth)
    )
    tif_shape = None
    tif_dtype = None
    with tifffile.TiffFile(all_tifs[0]) as im_tif:
        tif_shape = im_tif.pages[0].shape
        tif_dtype = im_tif.pages[0].dtype
    if not n_frames:
        n_frames = len(all_tifs)
    im_stack = [
        da.zeros(shape=tif_shape, dtype=tif_dtype) for _ in range(n_frames)
    ]

    @dask.delayed
    def read_im(tif_pth):
        with tifffile.TiffFile(tif_pth) as im_tif:
            im = im_tif.pages[0].asarray()
        return im

    for tif_pth in progress(all_tifs):
        frame_index = int(re.match(GT_TIF_REGEX, tif_pth).groups()[-1])
        im = da.from_delayed(
            read_im(tif_pth), shape=tif_shape, dtype=tif_dtype
        )
        im_stack[frame_index] = im
    layer_data = da.stack(im_stack)

    layer_type = "labels"
    layer_name = f"{gt_match.group(2)}{gt_match.group(3)}"
    add_kwargs = {"name": layer_name}

    return [(layer_data, add_kwargs, layer_type)]
